# js进阶知识 (上)

## 学习方法

1. 听明白, 理解所讲知识
2. 同组展开讨论, 弄明白没理解的
3. 用自己话进行总结, 并写在实训日志



## (一) 跨域

### (1) 跨域相关问题

#### 01 什么是跨域

**什么叫跨域**，浏览器有同源策略，只有当“协议”、“域名”、“端口号”都相同时，才能称之为是同源，其中有一个不同，即是跨域。

#### **02 同源策略的作用**是什么呢？

同源策略限制了从同一个源加载的文档或脚本如何与来自另一个源的资源进行交互。这是一个用于隔离潜在恶意文件的重要安全机制。

#### 03 我们又**为什么需要跨域**呢？

一是前端和服务器分开部署，接口请求需要跨域，二是我们可能会加载其它网站的页面作为iframe内嵌。

### **(2) 跨域的方法有哪些？**

1. jsonp跨域

2. cors跨域

   jsonp 只能支持 get 请求，cors 可以支持多种请求。cors是在服务器端设置, cors 并不需要前端做什么工作。

3. proxy(代理)跨域

   - nginx反向代理跨域
   - webpack代理跨域

4. websocket跨域

5. postMessage

6. 其它方式跨域

   - window.name + iframe
   - location.hash + iframe
   - document.domain 

### (3) 服务器端跨域

以nodejs的koa框架举例

```JavaScript
const cors = require('koa2-cors')
app.use(cors());
```

### (4) jsonp跨域

#### 01 jsonp问题

问题1: 为什么说jsonp不算真正的ajax?

1. ajax的核心是通过xmlHttpRequest获取非本页内容
   jsonp的核心是动态添加script标签调用服务器提供的js脚本
2. jsonp只支持get请求，ajax支持get和post请求

问题2: **jsonp跨域原理是什么?**

​		尽管浏览器有同源策略，但是 `<script>` 标签的 src 属性不会被同源策略所约束，可以获取任意服务器上的脚本并执行。jsonp 通过插入script标签的方式来实现跨域，参数只能通过url传入，仅能支持get请求。

#### **02 json跨域步骤**

**服务器端让接口支持jsonp**

```JavaScript
const jsonp = require('koa-jsonp');
app.use(jsonp());
```

**前端代码**

1. 创建用户接收数据的方法,比如getData
2. 添加script标签, src地址指向了jsonp接口地址,带上参数callback=getData

```html
<body>
    
    <script>
        function getData(data) {
            console.log(data);
        }
    </script>
    <script src="http://localhost:3000/user?callback=getData"></script>
</body>
```

注意事项: jsonp接口跟普通接口不太一样, 需要服务器端做特殊处理




## (二) 基本数据类型和引用数据类型的区别

1. 基本数据类型存放在栈区
2. 引用数据类型存放在堆区(它的引用存放在栈区)

   以下是基本数据类型和引用数据类型的存放方式

**牢记:**

* 变量(对象中的属性也是变量) => 容器 => 用来存放数据的

* 对象本体在内存中的地址, 以后统一称为引用(或者叫指针), 比如

  ```JavaScript
  // num是个容器, 存放了数字100
  // obj也是个容器, 存放的是对象的指针(引用)
  // obj.info也是一个容器, 存放的是{name:'张三'}的指针
  // 此刻, 内存中有2个对象
  let num = 100;
  const obj = {
      a: 2,
      b: 3,
      info: {
          name: '张三'
      }
  }
  ```
  
  ![image-20220928103955179](../../docs/img/image-20220928103955179.png)





练习: 说出下面代码运行的结果, 并说说为什么

```JavaScript
var a = 100;
var b = a;
b = 200;
console.log(a);  

var obj1 = {
	name: '张三',
  age: 100
}
var obj2 = obj1;
obj2.name = '李四';

console.log(obj1.name);   
```

```
const arr = [
	{name: '张三', age:100},
	{name: '李四', age:100},
    {name: '王五', age:100},
]

const obj = arr[2];
obj.age = 108;

console.log(arr[2].age);  // 108
```



## (三) 函数的调用方式(了解)

1. 普通调用

2. 回调

3. 递归调用

4. 自调用(尾调用)

   

   **回调函数**

   ```JavaScript
   // callback是个函数, 称为回调函数
   function foo(callback) {
     callback('hello');
   }
   
   function goo(msg) {
     console.log(msg);
   }
   
   // 调用foo, 把goo当做参数传入
   foo(goo);
   
   // 传入的函数是回调函数
   axios.get('xxxx').then(function(res) {
     console.log(res);
   }).catch(function(error) {
     console.log(error);
   });
   
   // 传入的函数是回调函数
   new Promise((resolve,reject) => {
       if('判断条件'){
         resolve();
       } else{
         reject();
       }
   })
   ```

   **回调函数, 函数自己调用自己**

   ```JavaScript
   // 使用setTimeout实现倒计时
   // 使用递归调用
   let count = 10;
   function countdown() {
     setTimeout(() => {
       console.log(count--);
       if (count >= 0) {
         countdown();
       }
     }, 1000);
   }
   countdown();
   ```

   **自调用(尾调用)**

   ```JavaScript
   (function () {
     var num = 100;
     console.log(num);
   })();
   
   (function(num){
     console.log(num);
   })(200)
   ```

   



## (四) js垃圾回收机制

#### (1) 垃圾回收机制是什么

垃圾回收机制(GC:Garbage Collection),执行环境负责管理代码执行过程中使用的内存。垃圾收集器会定期（周期性）找出那些不在继续使用的变量，然后释放其内存。但是这个过程不是实时的，因为其开销比较大，所以垃圾回收器会按照固定的时间间隔周期性的执行。

#### (2) 垃圾回收策略是什么

2种最为常用：标记清除和引用计数，其中标记清除更为常用。

**标记清除**(mark-and-sweep)：是对于脱离作用域的变量进行回收，当进入作用域时，进行标记，离开作用域时，标记并回收这些变量。到目前为止，IE、Firefox、Opera、Chrome、Safari的js实现使用的都是标记清除的垃圾回收策略或类似的策略，只不过垃圾收集的时间间隔互不相同。
         当变量进入环境（例如，在函数中声明一个变量）时，就将这个变量标记为“进入环境”。从逻辑上讲，永远不能释放进入环境的变量所占的内存，因为只要执行流进入相应的环境，就可能用到它们。而当变量离开环境时，这将其 标记为“离开环境”。 

**引用计数**：引用计数是跟踪记录每个值被引用的次数。就是变量的引用次数，被引用一次则加1，当这个引用计数为0时，被视为准备回收的对象，每当过一段时间开始垃圾回收的时候，就把被引用数为0的变量回收。引用计数方法可能导致循环引用，类似死锁，导致内存泄露。

#### (3) **如何减少垃圾回收开销**

由于每次的垃圾回收开销都相对较大，并且由于机制的一些不完善的地方，可能会导致内存泄露，我们可以利用一些方法减少垃圾回收，并且尽量避免循环引用。

1. 在对象结束使用后 ，令obj = null。这样利于解除循环引用，使得无用变量及时被回收；

2. js中开辟空间的操作有new(), [ ], { }, function (){..}。尽量减少此类操作, 最大限度的实现对象的重用；举例:

```JavaScript
var arr = [1,2,3];
// 若想清空数组, 你会怎么做
arr = [];  // 又创建了一个新的数组, 让arr指向了新的数组
arr.length = 0;
```

3. 慎用闭包。闭包容易引起内存泄露。本来在函数返回之后，之前的空间都会被回收。但是由于闭包可能保存着函数内部变量的引用，且闭包在外部环境，就会导致函数内部的变量不能够销毁。 

#### (4) 如何优化垃圾回收

分代回收（Generation GC）：与Java回收策略思想是一致的。目的是通过区分“临时”与“持久”对象；多回收“临时对象”区（young generation），少回收“持久对象”区（tenured generation），减少每次需遍历的对象，从而减少每次GC的耗时。
增量GC：这个方案的思想很简单，就是“每次处理一点，下次再处理一点，如此类推。 

## (五) 常见内存泄露的原因

- 全局变量引起的内存泄露
- 闭包引起的内存泄露：慎用闭包
- dom清空或删除时，事件未清除导致的内存泄漏
- 循环引用带来的内存泄露

参考连接:
https://blog.csdn.net/qq_21428081/article/details/82465801
https://www.cnblogs.com/kaicy/p/14715750.html





## (六) this的指向

问题：谈谈this的指向(背诵) 
或问：解释下JavaScript中this是如何工作

答: this的指向有以下几种情况

1. this永远指向函数运行时所在的对象，而不是函数被创建时所在的对象。
2. 普通的函数调用: 函数被谁调用，this就是谁。
3. 匿名函数或不处于任何对象中的函数指向window 。
4. 如果是call, apply, bind，this指向了这三者调用函数时传入的第一个参数
5. 箭头函数的指向: 由于箭头函数不绑定this， 它会捕获其所在上下文(作用域)的this值， 作为自己的this值

#### (1) 普通的函数调用

函数被谁调用，this就是谁。

```JavaScript
const obj = {
    say() {
        console.log(this);
    }
}

obj.say();  // this指向obj

const fn = obj.say;
window.fn();  // this指向window
```

#### (2) 匿名函数或不处于任何对象中的函数指向window

```JavaScript
 <script>
     const obj = {
         say: function() {
             console.log('say', this);
             setTimeout(function() {
                 console.log('setTimout',this)
             },0);
         }
     }

     obj.say();
 </script>
```

#### (3) call, apply, bind的指向

如果是call, apply, bind，this指向了这三者调用函数时传入的第一个参数, 详看 call, apply, bind下一个知识点:  call, apply和bind的区别

#### (4) 构造函数的this指向

构造函数的话，如果不用new操作符而直接调用，那即this指向window。用new操作符生成对象实例后，this就指向了新生成的对象。

```JavaScript
 <script>
     function Cat(name, age) {
         console.log('this', this);
         this.name = name;
         this.age = age;
     }
     // 直接调用, this指向window
     Cat('小花', 2);

     // 使用new, this就指向了创建出来的对象(实例)
     const cat = new Cat('小黑', '#000');
     console.log('cat',cat);
 </script>
```

#### (5) 箭头函数的this

由于箭头函数不绑定this， 它会捕获其所在上下文(所在作用域)的this值， 作为自己的this值

```javascript
<script>
    const obj = {
        say: function () {
            console.log('say', this);
            setTimeout(() => {
                console.log('setTimout', this)
            }, 0);
        }
    }

    obj.say();
</script>
```

## (七) call, apply和bind的区别

问题: call, apply和bind有什么区别

**相同点:**

​	三者都会改变this的指向, 

**不同点:**

call 和 apply 的功能相同，区别在于传参的方式不一样：

- call的第一个参数是this要指向的对象, 第二个,第三个....都是调用函数需要的参数。
- apply第一个参数是this要指向的对象, 第二个参数是数组或类数组, 数组是放的是调用函数需要的参数。

**例1. 改变this的指向**

```javascript
<script>
    var username = '张三';

    const obj1 = {
        username: '李四'
    }

    const obj2 = {
        username: '王五'
    }

    function say() {
        console.log(this, this.username);
    }

    console.log('直接调用的this')
    say(); // this指向window, username的知识张三

    // 调用call, this指向obj1
    console.log('call的this');
    say.call(obj1); 
</script>
```

**例2. call和apply的区别**

```JavaScript
<script>
    var username = '张三';

    const obj1 = {
        username: '李四'
    }

    const obj2 = {
        username: '王五'
    }

    function say(addr, age) {
        let str = `我叫${this.username},我${age}岁,我来自${addr}`;
        console.log(str);
    }
     
    // 使用让this指向obj1
    say.call(obj1,'广西',120);
    // 使用让this指向obj2,参数要放入数组
    say.apply(obj2,['广东',110]);

</script>
```



bind 和 call/apply 有一个很重要的区别，

- 一个函 数被 call/apply 的时候，会直接调用，但是 bind 会创建一个新函数, 不会直接调用

- 当这个新函数被调用时，bind() 的第一个参数将作为它运行时的 this, 新函数的调用和原来函数的调用一模一样,除了this的指向不一样之外

  ```JavaScript
  <script>
      var username = '张三';
      const obj = {
          username: '李四'
      }  
      function say(age) {
          console.log(this.username,age);
      } 
      const newSay = say.bind(obj);
      newSay(10); 
  </script>
  ```

  

## (八) 闭包及其作用

#### (1) 闭包是什么, 闭包形成的原因

**闭包的概念：**闭包是指有权访问另一个函数作用域中的变量的函数，创建闭包最常用的方式就是在一个函数内部创建另一个函数(函数中的函数)。

简单概括: 闭包就是一个能访问另外一个函数内部变量的函数

**闭包形成的原因：**函数内部的局部变量一般在函数运行结束的时候就会被销毁, 但如果局部变量因为被外部引用而导致没有被销毁, 就形成了闭包

**例1. 闭包应用1**

闭包导致内部变量被回收, 可以继续访问

```JavaScript
var obj = {
  init() {
    this.bindEvents();
  },

  bindEvents() {
    // 内部变量self, 因为被闭包使用, 所以不会被回收
    var self = this;
    // 点击删除
    $("#del").on("click", function () {
      self.del();
    });
  },

  del() {
    console.log('删除')
  }
};
```



**例2. 闭包的应用2**

闭包可以让我们在函数外部访问函数内部的变量

```JavaScript
function foo() {
  let count = 100; 

  return function() {
    return count;
  }

}
// 调用foo, 得到一个函数, 存入goo变量
var goo = foo();
// 调用goo, 就得到count的值
var count  = goo();
console.log('count', count);


// 另外一种写法
var count2 = foo()();
console.log('count2',count2);
```

#### (2) 闭包有哪些作用

或问: 闭包有哪些应用场景

1. 封装私有变量(私有属性)
2. 模仿块级作用域(ES5中没有块级作用域)
3. 实现JS的模块

##### 1.封装私有变量(私有属性)

私有变量(私有属性)是后台语言经常用的东西, 意思是这个属性是私有的,不能随意更改,要改,必须通过指定的set和get方法进行获取和修改

```JavaScript
function foo() {
  // 声明一个私有变量(跟普通变量是没有什么区别)
  var _name = "小花";
  // get方法用来获取_name变量的值
  function get() {
    return _name;
  }
  // set方法用来修改_name变量的值
  function set(newName) {
    _name = newName;
    console.log('新的名字是', newName);
  }

  return {
    get,
    set
  };
}

let obj = foo();
// 获取名字
let name = obj.get();
console.log('name',name);

// 修改名字
obj.set('小黑');
```



##### 2.模仿块级作用域

ES5中没有块级作用域

```javascript
// 期待隔一秒把i打印出来, i的取值是0,1,2,3,4
// 实际的结果是i全是5
for (var i = 0; i < 5; i++) {
  setTimeout(function () {
    console.log(i);
  }, i * 1000);
}

// 解决的办法1, 使用闭包, 模拟块级作用域
// 以下的代码里的自动用函数产生5个作用域, 每个作用的i都一样, 由于闭包的原因, i不会被回收, setTimout打印的时候就能读取得到
for (var i = 0; i < 5; i++) {
  (function (i) {
    setTimeout(function () {
      console.log(i);
    }, i * 1000);
  })(i);
}

// 解决的办法2, 使用es6的let是有块级作用域的
for (let i = 0; i < 5; i++) {
  setTimeout(function () {
    console.log(i);
  }, i * 1000);
}
```

##### 3.实现JS的模块(js库)

```javascript
(function () {
  var baseUrl = 'xxxx';

  function get() {
    // 具体的代码
    console.log(baseUrl)
  }

  function post() {
    // 具体的代码
    console.log(baseUrl)
  }

  window.axios = {
    get,
    post
  };
})();
```

##### **使用闭包的注意点**：

由于闭包会使得函数中的变量都被保存在内存中，内存消耗很大，所以不能滥用闭包，否则会造成网页的性能问题

## (九) 深拷贝和浅拷贝

#### 概念和定义

**定义:** 深拷贝和浅拷贝最根本的区别在于是否真正获取一个对象的复制实体，而不是引用。

**浅拷贝：**仅仅是指向被复制的内存地址，如果原对象发生改变，那么浅复制出来的对象也会相应的改变。

**深拷贝：**在计算机中开辟一块新的内存地址用于存放复制的对象。





#### -----------浅拷贝-----------

只拷贝引用

```JavaScript
<script>
    var obj = {
        name: '张三'
    } 
    var obj2 = obj;  // 浅拷贝
    obj.name = '李四';  // obj修改了name的值, obj2的name值也跟着发生改变
    console.log(obj2.name);
</script>
```



#### -----------深拷贝-----------

神拷贝又分为只拷贝第一层和完全拷贝

#### 1. 深拷贝方法 for in 的方法只能拷贝第一层

对象和数组都可以使用for in来进行深拷贝, 但只能拷贝第一层

```JavaScript
var obj1 = {
  username: '张三',
  age: 18,
  family: {
    num: 3,
    addr: '龙华'
  }
}

// 拷贝对象
var obj2 = {};
for(var key in obj1) {
  obj2[key] = obj1[key]; 
}
console.log(obj2);
console.log(obj2 === obj1); // 不是同一个对象
console.log(obj2.family === obj1.family); // 拷贝family的时候, 只拷贝了family的引用, 没有完全拷贝family对象



// 拷贝数组
var arr1 = ['a','b',{name: '张三'}]
var arr2 = [];
for(var key in arr1) {
  // console.log(arr1[key]);
  arr2[key] = arr1[key];
}

console.log(arr2 === arr1); // 得到了一个新的数组
console.log(arr2[2] === arr1[2]); // 引用类型只拷贝了引用,是同一个对象
```



#### 2. 使用es6的扩展运算符得到的对象也是只拷贝第一层

```JavaScript
var obj1 = {
  username: '张三',
  age: 18,
  family: {
    num: 3,
    addr: '龙华'
  }
} 

var obj2 = {...obj1};
console.log(obj1.family === obj2.family);
```

#### 3.  使用Object.assign()也是拷贝第一层

```JavaScript
var obj1 = {
  username: '张三',
  age: 18,
  family: {
    num: 3,
    addr: '龙华'
  }
}


var obj2 = Object.assign({},obj1);
console.log(obj1.family === obj2.family);
```

#### 4. 使用JSON的提供api

此方法是完全拷贝, 缺点: 函数拷贝不了, 会丢失

```javascript
var obj1 = {
  username: '张三',
  age: 18,
  family: {
    num: 3,
    addr: '龙华'
  },
  say() {
    console.log(this.username);
  }
}
 
// var str = JSON.stringify(obj1);
// var obj2 = JSON.parse(str);

// 合并上面两句代码
var obj2 = JSON.parse(JSON.stringify(obj1));


console.log(obj2.family === obj1.family);
console.log(obj2.say); // undefined
```

#### 5. 函数递归调用深拷贝 

**例1. 完全拷贝原理**

```javascript
<script>
    var obj = {
        a: 2,
        b: 3,
        cat: {
            name: "小花"
        }
    };

    var newObj = {};
    for (key in obj) {
        // 判断存放的数据类型是不是对象,
        // obj[key] !== null的时候,  obj[key]会自动转换成true, 所以 obj[key] !== null可以写成obj[key]或者!!obj[key]
        if (typeof obj[key] === "object" && obj[key] !== null) {
            // 相当于newObj.cat = {};
            newObj[key] = {};
            for (p in obj[key]) {
                newObj[key][p] = obj[key][p];
            }
        } else {
            newObj[key] = obj[key];
        }
    }
    console.log('对象完全拷贝')
    console.log(newObj === obj);
    console.log(newObj.cat === obj.cat);


    // 数组深拷贝
    var arr = [
        'a',
        'b',
        {
            name: '小花'
        }
    ];
    var newArr = [];
    for (key in arr) {
        if (typeof arr[key] === 'object' && arr[key]) {
            newArr[key] = {}
            for (p in arr[key]) {
                newArr[key][p] = arr[key][p];
            }
        } else {
            newArr[key] = arr[key];
        } 
    }
    console.log('数组完全拷贝')
    console.log(newArr === arr);
    console.log(newArr[2] === arr[2]);
</script>
```

**例2. 封装函数实现深拷贝**

```JavaScript
    function deepCopy(o) {
        var newO = Array.isArray(o)?[]:{};
        for(key in o) {
            // 判断是不是对象
            if (typeof o[key]==='object' && o[key]) {
                // 拷贝下一层
                newO[key] = deepCopy(o[key]);
            } else {
                newO[key] = o[key];
            }
        }
        return newO;
    }


    console.log('递归调用深拷贝')
    var obj = {
        a: 2,
        b: 3,
        cat: {
            name: "小花",
            addr: {
                msg: '美国'
            }
        }
    };
    var obj2 = deepCopy(obj); 
    console.log(obj2.cat.addr === obj.cat.addr); 
```



#### 其它方法:

https://www.jianshu.com/p/1c142ec2ca45

## (十) 严格模式'use strict'(了解)

 在代码中出现表达式-"use strict"; 意味着代码按照严格模式解析，这种模式使得Javascript在更严格的条件下运行。

**好处：**

1. 消除Javascript语法的一些不合理、不严谨之处，减少一些怪异行为;
2. 消除代码运行的一些不安全之处，保证代码运行的安全；
3. 提高编译器效率，增加运行速度；
4. 为未来新版本的Javascript做好铺垫。

**坏处：**同样的代码，在"严格模式"中，可能会有不一样的运行结果；

```js
<script>
     // 坏处
     "use strict";
     function say() {
         console.log(this);  // undefined
     }
     say();
</script>
```

```JavaScript
<script>
     // 好处
     "use strict"; 
     function say() {
         // 非严格模式, 不写var, 声明的变量是全局变量, 严格模式下不允许这么做
        a = 100;
     }
     say();
     console.log(a);
</script>
```



## (十一) 防抖和节流

#### 概念

在前端开发的过程中，我们经常会需要绑定一些持续触发的事件，如滚动、输入等等，但有些时候我们并不希望在事件持续触发的过程中那么频繁地去执行函数,防抖和节流是比较好的解决方案。

**(1)所谓防抖**，就是指触发事件后在 n 秒内函数只能执行一次，如果在 n 秒内又触发了事件，则会重新计算函数执行时间。使用场景:注册的时候检查用户名是否已经被注册.

**(2)所谓节流**，就是指连续触发事件但是在 n 秒中只执行一次函数。节流会稀释函数的执行频率。使用场景:监听滚动条是否到了顶部或底部.

#### (1) 防抖例子

**例1. 防抖的实现**

* 实现的原理是, 在输入的过程中, 清除上一次创建的延迟器
* 输入结束后, 最后一个延迟器没有被清除, 所以就可以被执行

```javascript
<body>

    <input type="text" oninput="handleInput()">

    <script>
        // 处理输入事件
        var timer;
        function handleInput() {
            // 输入的过程中, 不发请求
            clearTimeout(timer);
            timer = setTimeout(() => {
                doAjX();
            }, 1000);
        }

        function doAjX() {
            var url = 'http://81.71.65.4:3003/category/all';
            var xhr = new XMLHttpRequest();
            xhr.open('get', url);
            xhr.send();
            xhr.onreadystatechange = function () {
                if (xhr.readyState === 4 && xhr.status === 200) {
                    console.log('请求成功')
                }
            }
        }
    </script>
</body>
```

封装函数

```javascript
<body>

    <input type="text">

    <script>
        var _input = document.querySelector('input');
        // 绑定input事件
        _input.addEventListener('input', debounce(500, other), false)
        // 封装防抖函数
        function debounce(count, callback) {
            var timer;
            return function () {
                clearTimeout(timer);
                timer = setTimeout(() => {
                    // 执行回调函数
                    callback();
                }, count);
            }
        }

        function doAjax() {
            console.log('doAjax')
        };

        function other() {
            console.log('do other')
        }
    </script>
</body>
```

#### (2) 节流例子

**例1. 节流举例**

```javascript
<body>
    <div class="box">
        <p>ppppppppppppppppppppppppppppp</p>
        <p>ppppppppppppppppppppppppppppp</p>
        <p>ppppppppppppppppppppppppppppp</p>
        <p>ppppppppppppppppppppppppppppp</p>
        <p>ppppppppppppppppppppppppppppp</p>
        <p>ppppppppppppppppppppppppppppp</p> 
    </div>


    <script>  
        let count = 0;
        // 开始时间
        let startTime = Date.now();
        window.onscroll = function () {
            let endTime = Date.now();
            // 现在的时间和开始的时间进行比较
            if (endTime - startTime > 300) {
                console.log(count++);
                // 把endTime作为开始的时间重新进行计算
                startTime = endTime;
            } 
        }
    </script>
</body>
```

**例2. 封装函数**

```JavaScript
window.onscroll =  throttle(200,()=> {
    console.log(2222);
}); 

function throttle(count,callback) {
    let startTime = Date.now();
    return function () {
        let endTime = Date.now(); 
        if (endTime - startTime > count) {
            startTime = endTime;
            callback();
        }
    }
}
```



#### (3) 防抖和节流的区别

函数防抖和节流区别在于，当事件持续被触发，如果触发时间间隔短于规定的等待时间（n秒），那么

- 函数防抖的情况下，函数将一直推迟执行，造成不会被执行的效果；
- 函数节流的情况下， 函数将按指定的时间间隔持续执行。

## (十二) Eslint(了解)
http://docs.huruqing.cn/%E7%AC%AC12%E7%AB%A0%20%E5%B7%A5%E5%85%B7/VsCode.html#eslint%E9%85%8D%E7%BD%AE

### (1) Eslint常用命令 

团队合作时, 统一代码规范

1. 忽略 no-undef 检查 `/* eslint-disable no-undef */`

2. 忽略 no-new 检查 `/* eslint-disable no-new */`

3. 检查且修复 `eslint * --fix`

4. 检查指定文件 `eslint app.js --fix`

5. eslint初始化

   ```javascript
   npm init -y
   npx eslint --init
   ```

   

### (2) Eslint配置 

1. 安装eslint

   ```
   npm i eslint -g
   ```

   vscode搜索插件eslint安装

2. .eslintrc.js配置,使用`npx eslint --init`创建基本配置, 得到以下配置

   ```javascript
    module.exports = {
       "env": {
           "browser": true,
           "es2021": true
       },
       "extends": "eslint:recommended",
       "parserOptions": {
           "ecmaVersion": "latest",
           "sourceType": "module"
       },
       "rules": {
       }
   }
   ```

3. setting.json配置

   ```javascript
   {
       // 其它设置
       .....
       
       // eslint配置
       "eslint.alwaysShowStatus": true, // 显示eslint状态
       "eslint.format.enable": true,  // 使用eslint格式化代码
       "editor.codeActionsOnSave": {
           // 保存时自动修复所有错误
           "source.fixAll.eslint": true
       }
   }
   ```

4. 使用eslint规则格式化代码

   - 文件 -> 首选项 -> 扩展 -> eslint

     - eslint > format:Enable 把eslint作为一个格式化工具
     - 右键-> 使用...格式化文档 -> 配置默认格式化程序 -> 选择eslint
     - Eslint: always show status 在任务栏显示eslint状态

   - 保存时自动修复问题

     ```javascript
      "editor.codeActionsOnSave": {
         // 保存时自动修复所有错误
         "source.fixAll.eslint": true
       },
     ```

5. 自定义规则

   "off"或者0，不启用这个规则
   
   "warn"或者1，出现问题会有警告
   
   "error"或者2，出现问题会报错
   
   ```javascript
    'rules': {
           'no-alert': 0, //禁止使用alert confirm prompt
           'no-array-constructor': 2, //禁止使用数组构造器
           'no-bitwise': 0, //禁止使用按位运算符
           'no-caller': 1, //禁止使用arguments.caller或arguments.callee
           'no-catch-shadow': 2, //禁止catch子句参数与外部作用域变量同名
           'no-class-assign': 2, //禁止给类赋值
           'no-cond-assign': 2, //禁止在条件表达式中使用赋值语句
           'no-console': 0, //禁止使用console
           'no-const-assign': 2, //禁止修改const声明的变量
           'no-constant-condition': 2, //禁止在条件中使用常量表达式 if(true) if(1)
           'no-continue': 0, //禁止使用continue
           'no-control-regex': 2, //禁止在正则表达式中使用控制字符
           'no-debugger': 2, //禁止使用debugger
           'no-delete-var': 2, //不能对var声明的变量使用delete操作符
           'no-div-regex': 1, //不能使用看起来像除法的正则表达式/=foo/
           'no-dupe-keys': 2, //在创建对象字面量时不允许键重复 {a:1,a:1}
           'no-dupe-args': 2, //函数参数不能重复
           'no-duplicate-case': 2, //switch中的case标签不能重复
           'no-else-return': 2, //如果if语句里面有return,后面不能跟else语句
           'no-empty': 2, //块语句中的内容不能为空
           'no-empty-character-class': 2, //正则表达式中的[]内容不能为空
           'no-empty-label': 0, //禁止使用空label
           'no-eq-null': 2, //禁止对null使用==或!=运算符
           'no-eval': 1, //禁止使用eval
           'no-ex-assign': 2, //禁止给catch语句中的异常参数赋值
           'no-extend-native': 2, //禁止扩展native对象
           'no-extra-bind': 2, //禁止不必要的函数绑定
           'no-extra-boolean-cast': 2, //禁止不必要的bool转换
           'no-extra-parens': 2, //禁止非必要的括号
           'no-extra-semi': 2, //禁止多余的冒号
           'no-fallthrough': 1, //禁止switch穿透
           'no-floating-decimal': 2, //禁止省略浮点数中的0 .5 3.
           'no-func-assign': 2, //禁止重复的函数声明
           'no-implicit-coercion': 1, //禁止隐式转换
           'no-implied-eval': 2, //禁止使用隐式eval
           'no-inline-comments': 0, //禁止行内备注
           'no-inner-declarations': [2, 'functions'], //禁止在块语句中使用声明（变量或函数）
           'no-invalid-regexp': 2, //禁止无效的正则表达式
           'no-invalid-this': 2, //禁止无效的this，只能用在构造器，类，对象字面量
           'no-irregular-whitespace': 2, //不能有不规则的空格
           'no-iterator': 2, //禁止使用__iterator__ 属性
           'no-label-var': 2, //label名不能与var声明的变量名相同
           'no-labels': 2, //禁止标签声明
           'no-lone-blocks': 2, //禁止不必要的嵌套块
           'no-lonely-if': 2, //禁止else语句内只有if语句
           'no-loop-func': 1, //禁止在循环中使用函数（如果没有引用外部变量不形成闭包就可以）
           'no-mixed-requires': [0, false], //声明时不能混用声明类型
           'no-mixed-spaces-and-tabs': [2, false], //禁止混用tab和空格
           'linebreak-style': [0, 'windows'], //换行风格
           'no-multi-spaces': 1, //不能用多余的空格
           'no-multi-str': 2, //字符串不能用\换行
           'no-multiple-empty-lines': [1, { 'max': 2 }], //空行最多不能超过2行
           'no-native-reassign': 2, //不能重写native对象
           'no-negated-in-lhs': 2, //in 操作符的左边不能有!
           'no-nested-ternary': 0, //禁止使用嵌套的三目运算
           'no-new': 1, //禁止在使用new构造一个实例后不赋值
           'no-new-func': 1, //禁止使用new Function
           'no-new-object': 2, //禁止使用new Object()
           'no-new-require': 2, //禁止使用new require
           'no-new-wrappers': 2, //禁止使用new创建包装实例，new String new Boolean new Number
           'no-obj-calls': 2, //不能调用内置的全局对象，比如Math() JSON()
           'no-octal': 2, //禁止使用八进制数字
           'no-octal-escape': 2, //禁止使用八进制转义序列
           'no-param-reassign': 2, //禁止给参数重新赋值
           'no-path-concat': 0, //node中不能使用__dirname或__filename做路径拼接
           'no-plusplus': 0, //禁止使用++，--
           'no-process-env': 0, //禁止使用process.env
           'no-process-exit': 0, //禁止使用process.exit()
           'no-proto': 2, //禁止使用__proto__属性
           'no-redeclare': 2, //禁止重复声明变量
           'no-regex-spaces': 2, //禁止在正则表达式字面量中使用多个空格 /foo bar/
           'no-restricted-modules': 0, //如果禁用了指定模块，使用就会报错
           'no-return-assign': 1, //return 语句中不能有赋值表达式
           'no-script-url': 0, //禁止使用javascript:void(0)
           'no-self-compare': 2, //不能比较自身
           'no-sequences': 0, //禁止使用逗号运算符
           'no-shadow': 2, //外部作用域中的变量不能与它所包含的作用域中的变量或参数同名
           'no-shadow-restricted-names': 2, //严格模式中规定的限制标识符不能作为声明时的变量名使用
           'no-spaced-func': 2, //函数调用时 函数名与()之间不能有空格
           'no-sparse-arrays': 2, //禁止稀疏数组， [1,,2]
           'no-sync': 0, //nodejs 禁止同步方法
           'no-ternary': 0, //禁止使用三目运算符
           'no-trailing-spaces': 1, //一行结束后面不要有空格
           'no-this-before-super': 0, //在调用super()之前不能使用this或super
           'no-throw-literal': 2, //禁止抛出字面量错误 throw "error";
           'no-undef': 1, //不能有未定义的变量
           'no-undef-init': 2, //变量初始化时不能直接给它赋值为undefined
           'no-undefined': 2, //不能使用undefined
           'no-unexpected-multiline': 2, //避免多行表达式
           'no-underscore-dangle': 1, //标识符不能以_开头或结尾
           'no-unneeded-ternary': 2, //禁止不必要的嵌套 var isYes = answer === 1 ? true : false;
           'no-unreachable': 2, //不能有无法执行的代码
           'no-unused-expressions': 2, //禁止无用的表达式
           'no-unused-vars': [2, { 'vars': 'all', 'args': 'after-used' }], //不能有声明后未被使用的变量或参数
           'no-use-before-define': 2, //未定义前不能使用
           'no-useless-call': 2, //禁止不必要的call和apply
           'no-void': 2, //禁用void操作符
           'no-var': 0, //禁用var，用let和const代替
           'no-warning-comments': [
               1,
               { 'terms': ['todo', 'fixme', 'xxx'], 'location': 'start' }
           ], //不能有警告备注
           'no-with': 2, //禁用with
   
           'array-bracket-spacing': [2, 'never'], //是否允许非空数组里面有多余的空格
           'arrow-parens': 0, //箭头函数用小括号括起来
           'arrow-spacing': 0, //=>的前/后括号
           'accessor-pairs': 0, //在对象中使用getter/setter
           'block-scoped-var': 0, //块语句中使用var
           'brace-style': [1, '1tbs'], //大括号风格
           'callback-return': 1, //避免多次调用回调什么的
           'camelcase': 2, //强制驼峰法命名
           'comma-dangle': [2, 'never'], //对象字面量项尾不能有逗号
           'comma-spacing': 0, //逗号前后的空格
           'comma-style': [2, 'last'], //逗号风格，换行时在行首还是行尾
           'complexity': [0, 11], //循环复杂度
           'computed-property-spacing': [0, 'never'], //是否允许计算后的键名什么的
           'consistent-return': 0, //return 后面是否允许省略
           'consistent-this': [2, 'that'], //this别名
           'constructor-super': 0, //非派生类不能调用super，派生类必须调用super
           'curly': [2, 'all'], //必须使用 if(){} 中的{}
           'default-case': 2, //switch语句最后必须有default
           'dot-location': 0, //对象访问符的位置，换行的时候在行首还是行尾
           'dot-notation': [0, { 'allowKeywords': true }], //避免不必要的方括号
           'eol-last': 0, //文件以单一的换行符结束
           'eqeqeq': 2, //必须使用全等
           'func-names': 0, //函数表达式必须有名字
           'func-style': [0, 'declaration'], //函数风格，规定只能使用函数声明/函数表达式
           'generator-star-spacing': 0, //生成器函数*的前后空格
           'guard-for-in': 0, //for in循环要用if语句过滤
           'handle-callback-err': 0, //nodejs 处理错误
           'id-length': 0, //变量名长度
           'indent': [2, 4], //缩进风格
           'init-declarations': 0, //声明时必须赋初值
           'key-spacing': [0, { 'beforeColon': false, 'afterColon': true }], //对象字面量中冒号的前后空格
           'lines-around-comment': 0, //行前/行后备注
           'max-depth': [0, 4], //嵌套块深度
           'max-len': [0, 80, 4], //字符串最大长度
           'max-nested-callbacks': [0, 2], //回调嵌套深度
           'max-params': [0, 3], //函数最多只能有3个参数
           'max-statements': [0, 10], //函数内最多有几个声明
           'new-cap': 2, //函数名首行大写必须使用new方式调用，首行小写必须用不带new方式调用
           'new-parens': 2, //new时必须加小括号
           'newline-after-var': 2, //变量声明后是否需要空一行
           'object-curly-spacing': [0, 'never'], //大括号内是否允许不必要的空格
           'object-shorthand': 0, //强制对象字面量缩写语法
           'one-var': 1, //连续声明
           'operator-assignment': [0, 'always'], //赋值运算符 += -=什么的
           'operator-linebreak': [2, 'after'], //换行时运算符在行尾还是行首
           'padded-blocks': 0, //块语句内行首行尾是否要空行
           'prefer-const': 0, //首选const
           'prefer-spread': 0, //首选展开运算
           'prefer-reflect': 0, //首选Reflect的方法
           'quotes': [1, 'single'], //引号类型 `` "" ''
           'quote-props': [2, 'always'], //对象字面量中的属性名是否强制双引号
           'radix': 2, //parseInt必须指定第二个参数
           'id-match': 0, //命名检测
           'require-yield': 0, //生成器函数必须有yield
           'semi': [2, 'always'], //语句强制分号结尾
           'semi-spacing': [0, { 'before': false, 'after': true }], //分号前后空格
           'sort-vars': 0, //变量声明时排序
           'space-after-keywords': [0, 'always'], //关键字后面是否要空一格
           'space-before-blocks': [0, 'always'], //不以新行开始的块{前面要不要有空格
           'space-before-function-paren': [0, 'always'], //函数定义时括号前面要不要有空格
           'space-in-parens': [0, 'never'], //小括号里面要不要有空格
           'space-infix-ops': 0, //中缀操作符周围要不要有空格
           'space-return-throw-case': 0, //return throw case后面要不要加空格
           'space-unary-ops': [0, { 'words': true, 'nonwords': false }], //一元运算符的前/后要不要加空格
           'spaced-comment': 0, //注释风格要不要有空格什么的
           'strict': 2, //使用严格模式
           'use-isnan': 2, //禁止比较时使用NaN，只能用isNaN()
           'valid-jsdoc': 0, //jsdoc规则
           'valid-typeof': 2, //必须使用合法的typeof的值
           'vars-on-top': 2, //var必须放在作用域顶部
           'wrap-iife': [2, 'inside'], //立即执行函数表达式的小括号风格
           'wrap-regex': 0, //正则表达式字面量用小括号包起来
           'yoda': [2, 'never'] //禁止尤达条件
       }
   ```



## (十三) 队列和栈 

队列和栈都是后台语言的数据结构, js可以通过数组的一些方法进行模拟队列和栈

#### 栈 

1. 定义: 栈是一种后进先出的数据结构，也就是说最新添加(后进)的项最早被移出；它是一种运算受限的线性表，只能在表头/栈顶进行插入和删除操作。栈有栈底和栈顶。

2. 入栈和出栈: 向一个栈插入新元素叫入栈（进栈），就是把新元素放入到栈顶的上面，成为新的栈顶；从一个栈删除元素叫出栈，就是把栈顶的元素删除掉，相邻的成为新栈顶；

3. js数组方法模拟栈的数据结构
   1. 入栈 push()
   2. 出栈 pop()
   
   ```JavaScript
   // 栈操作
   var arr = ['a','b','c'];
   arr.push('d');  // 入栈
   arr.pop(); // 出栈
   ```

#### 队列

1. 定义: 队列是一种先进先出的数据结构。 队列在列表的末端增加项，在首端移除项。

2. 出队列和入队列：队列允许在表的首端（队列头）进行删除操作，在表的末端（队列尾）进行插入操作；（例如在银行窗口排队办理业务，最前面的第一个人开始办理，后面来的人只能在队伍末尾排队，直到排到他们为止），队列是实现多任务的重要机制！

3. js数组模拟入队列和出队列
   - 入队列 unshift()
   - 出队列 pop()
   
   ```JavaScript
   // 队列操作
   var arr2 = ['a','b','c'];
   arr2.unshift(1);  // 入队列
   arr2.pop(); // 出队列
   ```
   
   

## (十四) 宏任务和微任务

### --------核心知识点--------

### (1) JavaScript线程

众所周知，JS 是一门单线程语言，可是浏览器又能很好的处理异步请求，那么到底是为什么呢？(单线程意思是一次只能处理一件事情, 就好像一条小马路, 一次只能一辆车通过)

JS 的执行环境一般是浏览器和 Node.js，两者稍有不同，这里只讨论浏览器环境下的情况。

```
(1) 所有同步任务都在主线程上执行，形成一个执行栈
(2) 主线程之外，还存在一个"任务队列", 用来处理异步任务, 异步任务一般通过函调函数来进行执行
(3) 执行栈"中的所有同步任务执行完毕，系统就会读取"任务队列"并执行
```

### (2) js的两种任务

JS 执行过程中会产生两种任务，分别是：同步任务和异步任务。

- 同步任务：比如声明语句、for、赋值等，读取后依据从上到下从左到右，立即执行。
- 异步任务：比如 ajax 网络请求，setTimeout 定时函数等都属于异步任务。异步任务会通过任务队列(**Event Queue**)的机制（先进先出的机制）来进行协调。

### (3) **任务队列（Event Queue）**

任务队列: 也就是有很多任务, 所以需要把任务排个队, 然后按一定的规则执行任务

任务队列中的任务也分为两种，分别是：宏任务（Macro-take）和微任务（Micro-take）

- 宏任务主要包括：scrip(JS 整体代码)、setTimeout、setInterval、setImmediate、I/O、UI 交互
- 微任务主要包括：Promise(重点关注)、process.nextTick(Node.js)、MutaionObserver

任务队列的执行过程是：**先执行一个宏任务**，执行过程中如果产出新的宏/微任务，就将他们推入相应的任务队列，**之后再执行一队微任务(把微任务队列所有任务执行完)**，之后再执行宏任务，如此循环。**以上不断重复的过程就叫做 Event Loop(事件循环)**。

每一个宏任务称为一个task, 每一次的循环操作被称为**tick**。

![图片](../../docs/img/640.png)



### -------例子和延伸------

### (1) 宏任务和微任务的执行过程

宏任务和微任务的执行过程也称为事件循环机制, 执行顺序就是: 宏任务 => 微任务 => 宏任务 => 微任务 => ......

#### **例1. setTimout和ajax**

```JavaScript
<script src="https://cdn.bootcdn.net/ajax/libs/axios/0.27.2/axios.min.js"></script>
<script> 
  console.log('start'); 
  setTimeout(() => {
    console.log('setTimeout');
  }, 0)

  var xhr = new XMLHttpRequest();
  xhr.open('get', 'http://81.71.65.4:3003/category/all');
  xhr.send();
  xhr.onreadystatechange = function () {
    if (xhr.status === 200 && xhr.readyState === 4) {
      console.log('收到数据');
    }
  }

  console.log('end');
</script>
```

**解析:** 

一开始宏任务队列只有一个[script]

**(1) 执行第1个宏任务**

1. 打印start
2. 产生一个宏任务,宏任务队列[setTimeout]
3. 又产生了另一个宏任务, 此时宏任务队列[setTimeout, ajax]
4. 打印end

​    **注意**: setTimeout或者ajax任务, 当它们准备好的时候才推入任务队列

**(2) 执行第2个宏任务**

1. 打印了setTimeout

**(3) 执行第3个宏任务**

1. 打印收到数据





#### **例2. promise举例(上)**

`new Promise(..)`中的代码，也是同步代码，会立即执行。只有`then`之后的代码，才是异步执行的代码，是一个微任务。

```JavaScript
console.log("script start");

setTimeout(function () {
    console.log("setTimeout");
}, 0);

const promise = new Promise((resolve, reject) => {
    resolve('哈哈哈');
})
promise.then(() => {
    console.log("promise");
})
console.log("script end");
```

**解析:** 

**(1) 执行第1个宏任务[script]**

1. 打印script start
2. 把宏任务setTimeout推入宏任务队列 [setTimeout]
3. 把.then 推入微任务队列 [then]
4. 打印script end

**(2) 执行微任务队列的所有任务**

1. 打印promise

**(3) 继续执行第2个宏任务**

1. 打印setTimeout

运行结果:

```
script start
script end
promise
setTimeout 
```



#### **例3. promise举例(中)**

* .then()方法返回的也是promise对象, 还可以继续.then()
* .then()是微任务, 执行微任务的过程中, 产生微任务,就放入微任务队列, 继续执行, 直到队列没有任务为止

例3.1 .then的例子

```JavaScript
var promise = new Promise((resolve)=> {
  resolve('hello');
});

// 另一种创建promise对象的方法
// var promise = Promise.resolve('hello');

promise.then(res=> {
  return res+1;
}).then(res=> {
  return res+1;
}).then(res=> {
  console.log(res);
}) 
```

例3.2 promise举例分析

```javascript
console.log("script start");

setTimeout(function () {
    console.log("setTimeout");
}, 0);

const promise = new Promise((resolve, reject) => {
    resolve('promise');
}) 
promise.then(res=> {
    console.log('then1');
}).then(res=> {
    console.log('then2');
}).then(res=> {
    console.log('then3');
})

console.log("script end");
```

**解析:**

**(1) 执行第1个宏任务[script]**

1. 打印script start
2. 把宏任务推入宏任务队里 [setTimeout]
3. .then() 产生微任务then1, 微任务队列[then1]
4. 打印script end

**(2) 执行微任务队列里的所有任务**

1. 执行then1, 打印'then1'
2. 产生新的微任务[then2]
3. 执行then2, 打印'then2'
4. 产生新的微任务[then3]
5. 执行then3, 打印'then3'

**(3) 执行宏任务[setTimeout]**

1. 打印setTimeout



例3.3 async await

```JavaScript
const promise = new Promise((resolve, reject) => {
    resolve('hello');
})   

// async function say() {
//   console.log(1)
//   let res = await promise;
//   console.log(res);
//   console.log(2);
// }
// say();


function say2() {
  console.log(1);
  promise.then(res=> {
    console.log(res);
  }).then(res=> {
    console.log(2);
  })
}
say2();
```



**例4. promise举例(下)**

```JavaScript
console.log("script start");  

setTimeout(function () {
  console.log("timeout1"); 
}, 10);

new Promise((resolve) => {
  console.log("promise1");  
  resolve();
  setTimeout(() => console.log("timeout2"), 0);  
}).then(function () {
  console.log("then1");  
});

console.log("script end");  
```

**解析:**

**(1) 执行第1个宏任务[script]**

1. 打印script start
2. 把timeout1推入宏任务队列[timeout1]
3. 打印promise1
4. 把timeout2推入宏任务队列[timeout1,timeout2]
5. 把微任务then1推入微任务队列[then1]
6. 打印script end

**(2) 微任务[then1]**

1. 打印then1

**(3) 执行宏任务**[timeout1,timeout2], timeout2先执行

1. 打印timeout2

**(4)执行宏任务[timeout1]** 

1. 打印timeout1

输出结果:

```
script start
promise1
script end
then1
timeout2
timeout1
```



### **(2) setTimerout 并不准确**

由上我们已经知道了 setTimeout 是一个宏任务，会被添加到宏任务队列当中去，按顺序执行。

**setTimeout() 的第二个参数是为了告诉 JavaScript 再过多长时间把当前任务添加到队列中。**

如果队列是空的，那么添加的代码会立即执行；如果队列不是空的，那么它就要等前面的代码执行完了以后再执行。

看代码：

```javascript
const s = new Date().getSeconds();
console.log("script start");
setTimeout(() => {
  console.log("timeout");
}, 2000);
new Promise((resolve) => {
  console.log("promise");
  resolve();
}).then(() => {
    // 开始时间
    let start = new Date();
    // 结束时间
    let end;
    for(;;) {
        end = new Date();
        // 结束时间比开始时间多3秒, 就跳出
        if (end.getSeconds() - start.getSeconds() > 10) {
            break;
        }
    }
    console.log('10秒之后');
});
console.log("script end");
```

**解析:**



### **(3) 总结**

1. "任务队列"是一个事件的队列（也可以理解成消息的队列），IO设备完成一项任务，就在"任务队列"中添加一个事件，表示相关的异步任务可以进入"执行栈"了。主线程读取"任务队列"，就是读取里面有哪些事件。

2. "任务队列"中的事件，除了IO设备的事件以外，还包括一些用户产生的事件（比如鼠标点击、页面滚动等等）。只要指定过回调函数，这些事件发生时就会进入"任务队列"，等待主线程读取。所谓"回调函数"（callback），就是那些会被主线程挂起来的代码。异步任务必须指定回调函数，当主线程开始执行异步任务，就是执行对应的回调函数。

3. "任务队列"是一个先进先出的数据结构，排在前面的事件，优先被主线程读取。主线程的读取过程基本上是自动的，只要执行栈一清空，"任务队列"上第一位的事件就自动进入主线程。

4. 从规范来看，microtask(微任务) 优先于 task(宏任务) 执行，所以如果有需要优先执行的逻辑，放入 microtask 队列会比 task 更早的被执行。

5. JavaScript 是一门单线程语言，异步操作都是放到事件循环队列里面，等待主线程来执行的，并没有专门的异步执行线程(html5倒是提供了一个异步线程webworker, 但这是浏览器提供)。



## (十五) 算法 

参考链接:  https://www.cnblogs.com/AlbertP/p/10847627.html

数组里的sort方法采用的冒泡排序

![img](http://zk.huruqing.cn/assets/2.6829c1c1.png)

```javascript
<script>
     // i=0; 第1次循环, 会把最大放到了最后, 两相比4次  5-0-1 = 4
     // i=1; 第2次循环, 会把第2大放到了最后, 两两相比3次 5-1-1 = 3
	 // 所以第二个for循环的次数是length-i-1
    
    function bubbleSort(arr) { 
    var len = arr.length;
    for (var i = 0; i < len; i++) { 
        for (var j = 0; j < len - i; j++) {
            if (arr[j] > arr[j + 1]) { //相邻元素两两对比
                var temp = arr[j + 1]; //元素交换
                arr[j + 1] = arr[j];
                arr[j] = temp;
            }
        } 
    }
    return arr;
} 

var arr = [8,2,9,1,7];
var newArr = bubbleSort(arr); 
</script> 
```

![img](http://zk.huruqing.cn/assets/strip.287f9a70.gif)



排序演示代码

```html
<!DOCTYPE html>
<head>
    <meta charset="UTF-8">
    <style>
        .box { margin-top: 50px; border-bottom: 1px solid; display: flex; align-items: flex-end; width: 500px;}
        p { width: 40px; text-align: center; display: flex; align-items: center; justify-content: center; }
        p+p {margin-left: 20px;}
        .p1 { background-color: gray; height: 160px; }
        .p2 { background-color: gold; height: 40px; }
        .p3 { background-color: green; height: 180px; }
        .p4 { background-color: blue; height: 20px; }
        .p5 { background-color: pink; height: 140px; }
        .p6 { background-color: salmon; height: 60px; }
        .p7 { background-color: rgb(49, 199, 86); height: 120px; }
        .p8 { background-color: rgb(250, 114, 155); height: 80px; }
        .p9 { background-color: rgb(98, 62, 228); height: 100px; }
    </style>
</head>

<body>
    <div class="box">
        <p class="p1">8</p>
        <p class="p2">2</p>
        <p class="p3">9</p>
        <p class="p4">1</p>
        <p class="p5">7</p>
        <p class="p6">3</p>
        <p class="p7">6</p>
        <p class="p8">4</p>
        <p class="p9">5</p>
    </div>
</body> 
</html>
```



## (十六) js设计模式

假设有一个空房间，我们要日复一日地往里 面放一些东西。最简单的办法当然是把这些东西 直接扔进去，但是时间久了，就会发现很难从这 个房子里找到自己想要的东西，要调整某几样东 西的位置也不容易。所以在房间里做一些柜子也 许是个更好的选择，虽然柜子会增加我们的成 本，但它可以在维护阶段为我们带来好处。使用 这些柜子存放东西的规则，或许就是一种模式。

### (1) 单例模式

单例就是保证一个类只有一个实例，实现的方法一般是先判断实例存在与否，如果存在直接返回，如果不存在就创建了再返回，这就确保了一个类只有一个实例对象好处是可以减少不必要的内存开销 

应用场景:

* 数据库连接对象
* 网站的登录弹窗，无论在别的地方点击多少次，都只会弹出一个登录弹窗
* 全局数据存储对象（vuex）

```JavaScript
// 数据库连接对象
var MongoClient = require("mongodb").MongoClient;
// huruqing 数据库名称
let dbo;
async function createDb() {
  try {
    if (!dbo) {
      var url = "mongodb://localhost:27017/huruqing";
      var primise = await MongoClient.connect(url);
      // dbo就是操作数据库的一个对象
      dbo = primise.db("huruqing");
      console.log("数据库连接成功");
      console.log('创建新的数据库连接对象')
      return dbo;
    } else {
      console.log('使用原来的数据库连接对象')
      return dbo;
    }
  } catch (error) {
    console.log("错误信息:", error);
  }
}
module.exports = createDb;
```



### (2) 工厂模式

1. 工厂模式是一种用来创建对象的设计模式。

2. 我们不暴露对象创建的逻辑,而是将逻辑封装在一个函数内,那么这个函数可以称为工厂。

3. 作用(应用场景): 用来创建相似对象时执行一些重复操作。

   ```javascript
   function madeCat(name, age) {
     var obj = {};
     obj.name = name;
     obj.age = age;
     obj.say = () => {
       console.log(this.name);
     };
     return obj;
   }
   
   var cat1 = madeCat("小花", 1);
   var cat2 = madeCat("小白", 2);
   console.log(cat1);
   console.log(cat2);
   ```

**优缺点**

**优点**：通过工厂模式，我们可以快速创建大量相似对象，没有重复代码。
**缺点**： 

1. 工厂模式创建的对象属于Object，无法区分对象类型，这也是工厂模式没有广泛使用的原因
2. 每个对象都有自己的函数(像下面的例子中, 每创建一个对象, 就会创建两个函数, 若创建10个对象,就需要创建20个函数, 而这些函数都一模一样的, 完全没必要)，浪费资源

```js
<script>
    function madeCat(name, age) {
        var obj = {};
        obj.name = name;
        obj.age = age;
        obj.say = () => {
            console.log(this.name);
        };
        return obj;
    }

    var cat1 = madeCat("小花", 1);  
    console.log('cat1',cat1);
 
    var cat2 = madeCat("小黑", 1);  
    // 创建两个小猫, 就会创建两个方法
    console.log(cat1.say === cat2.say);
</script>
```

###  (3) 发布-订阅模式

发布---订阅模式又叫观察者模式，它定义了对象间的一种一对多的关系，让多个观察者对象同时监听某一个主题对象，当一个对象发生改变时，所有依赖于它的对象都将得到通知。

现实生活中的发布-订阅模式；

```JavaScript
比如小红最近在淘宝网上看上一双鞋子，但是呢 联系到卖家后，才发现这双鞋卖光了，但是小红对这双鞋又非常喜欢，所以呢联系卖家，问卖家什么时候有货，卖家告诉她，要等一个星期后才有货，卖家告诉小红，要是你喜欢的话，你可以收藏我们的店铺，等有货的时候再通知你，所以小红收藏了此店铺，但与此同时，小明，小花等也喜欢这双鞋，也收藏了该店铺；等来货的时候就依次会通知他们；

在上面的故事中，可以看出是一个典型的发布订阅模式，卖家是属于发布者，小红，小明等属于订阅者，订阅该店铺，卖家作为发布者，当鞋子到了的时候，会依次通知小明，小红等，依次使用旺旺等工具给他们发布消息；
```

##### 发布订阅模式的优点：

1.支持简单的广播通信，当对象状态发生改变时，会自动通知已经订阅过的对象。
2.比如上面的列子，小明，小红不需要天天逛淘宝网看鞋子到了没有，在合适的时间点，发布者(卖家)来货了的时候，会通知该订阅者(小红，小明等人)。

```tex
对于第一点，我们日常工作中也经常使用到，比如我们的ajax请求，请求有成功(success)和失败(error)的回调函数，我们可以订阅ajax的success和error事件。我们并不关心对象在异步运行的状态，我们只关心success的时候或者error的时候我们要做点我们自己的事情就可以了~
```

##### 例子1: js的事件

js的事件就是发布-订阅模式

```html
<body>
    <button>发布消息</button>

    <script>
        var button = document.querySelector('button');
        button.addEventListener('click', reader1, false);
        button.addEventListener('click', reader2, false);

        // 读者1
        function reader1() {
            console.log('读者1收到了信息');
        }
        // 读者2
        function reader2() {
            console.log('读者2收到了信息');
        }
    </script>
</body>
```



##### 例2. 用js实现发布-订阅模式

```javascript
<!DOCTYPE html>
<html lang="en">

<head>
    <meta charset="UTF-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Document</title>
</head>

<body>
    <select>
        <option label="深圳日报" value="szrb"></option>
        <option label="广州日报" value="gzrb"></option>
        <option label="南方日报" value="nfrb"></option>
    </select>
    <button>发布报纸</button>

    <script>
        // 报社的构造函数
        function Office() {

            // 存放读者订阅报纸的信息
            this.info = {
                'szrb': [],
                'gzrb': [],
                'nfrb': [],
            };

            // 发布报纸的功能
            this.publish = (newspaper) => { 
                // 获取报纸对应的读者
                var readerList = this.info[newspaper];
                // 通知数组的所有读者
                readerList.forEach(reader => {
                    // 报纸信息
                    var paperInfo;
                    switch (newspaper) {
                        case 'szrb':
                            paperInfo = {
                                [newspaper]: '深圳日报',
                                msg: '这是2022年9月28日的报纸'
                            }
                            break;
                        case 'gzrb':
                            paperInfo = {
                                [newspaper]: '广州日报',
                                msg: '这是2022年9月28日的报纸'
                            }
                            break;
                        case 'nfrb':
                            paperInfo = {
                                [newspaper]: '南方日报',
                                msg: '这是2022年9月28日的报纸'
                            }
                            break;
                    } 
                    // 把报纸信息传给读者
                    reader(paperInfo);
                })
            }

            /**
             * 订阅报纸
             * 参数1: 报纸的名称 
             * 参数2: 读者
             */
            this.subscribe = (newspaper, reader) => {
                // 把信息存入info
                this.info[newspaper].push(reader);
            }
        }

        var office = new Office();


        // 读者1订阅了深圳日报
        function reader1(data) {
            console.log(data)
        }
        office.subscribe('szrb', reader1);


        // 读者2订阅了深圳日报
        function reader2(data) {
            console.log(data)
        }
        office.subscribe('szrb', reader2);

        // 读者3订阅了深圳日报
        function reader3(data) {
            console.log(data)
        }
        office.subscribe('gzrb', reader3);
        console.log(office.info);


        document.querySelector('button').addEventListener('click', () => {
            var newspaper = document.querySelector('select').value;
            office.publish(newspaper);
        }, false)
    </script>
</body> 
</html>
```





## (十七) 手写promise

### (1) promise的用法

* 编写一个构造函数MyPromise
* 实现.then的功能
* 实现.catch的功能
* 实现多个.then功能, 链式调用

```javascript
const promise = new Promise((resolve,reject)=> {
    var isLogin = false;
    if (isLogin) {
        resolve('已登录')
    } else {
         reject('还没登录')
    } 
})

promise.then(res=>{
    console.log(res);
    return res+1;
}).then(res=> {
    console.log(res);
}).catch(err=> {
    console.log(err);
})
```

### (2) promise的实现

以下是一个简单的promise实现, 想更深一步研究, 请看末尾的链接

**例1. resolve和then方法的实现**

```javascript
<script>
    // 声明一个构造函数
    function MyPromise(callback) {
        // 2. 执行构造函数来创建对象
        console.log('执行构造函数')
        // 3. 定义sucessInfo来存放成功的数据
        this.sucessInfo = null;
        // 4. 定义then方法, 提供给实例来调用
        this.then = (callback) => {
            // 11.执行调用这传过来的函数, 并把this.sucessInfo数据传入给对方
            console.log("then方法");
            callback(this.sucessInfo);
        };

        // 5.定义resolve方法,用来保存成功的数据
        this.resolve = (data) => {
            // 9.把数据传入到this.sucessInfo
            console.log("在promise里调用resolve方法保存数据");
            this.sucessInfo = data;
        };
        // 6. 执行创建对象时传过来的函数
        callback(this.resolve);
    }

    // 1.创建一个promise对象, 传入一个函数
    const promise = new MyPromise((resolve) => {
        // 7. 接收Promise传过来的resolve方法
        console.log('创建promise对象');
        // 8. 调用resolve方法
        resolve(1111);
    });

    // 10.调用对象的then方法
    promise.then((res) => {
        console.log(res); // 打印1111
    });
</script>
```

**例2. 实现.catch方法, 处理失败的情况**

```JavaScript
<script>
    function MyPromise(callback) {
        // 存放成功的信息
        this.sucessInfo = null;
        // 存放失败失败的信息
        this.errorInfo = null;

        // 处理成功的信息
        this.then = (callback) => {
            callback(this.sucessInfo);
        };
        // 处理失败的信息
        this.catch = (callback) => {
            callback(this.errorInfo);
        };

        // 将成功信息保存到sucessInfo
        this.resolve = (data) => {
            this.sucessInfo = data;
        };
        // 将失败信息保存到errorInfo
        this.reject = (data) => {
            this.errorInfo = data;
        };

        callback(this.resolve, this.reject);
    }

    const promise = new MyPromise((resolve, reject) => {
        let flag = false;
        if (flag) {
            resolve('我成功了, 哈哈哈哈哈哈');
        } else {
            reject('失败了, 继续努力');
        }
    });


    promise.then((res) => {
        console.log(res);
    })
    promise.catch((err) => {
        console.log(err);
    })
</script>
```

**例3. 链式调用的原理**

调用对象方法的时候, 把对象返回即可

```JavaScript
function Cat(name, age) {
  this.name = name;
  this.age = age;
  this.sayName = () => {
    console.log(this.name);
    return this;
  };
  this.sayAge = () => {
    console.log(this.age);
    return this;
  };
}

const cat = new Cat("小花", 1);
 
cat.sayName().sayAge();
```

**例4. 实现promise的链式调用**

* 调用then或者catch时把this返回
* 根据状态执行响应的回调, 成功就执行成功的回调, 失败就执行失败的回调

```JavaScript
<script>
    function MyPromise(callback) {
        // 存放成功的信息
        this.sucessInfo = null;
        // 存放失败失败的信息
        this.errorInfo = null;
        // 定义promise的状态
        this.status = 'pending';

        // 处理成功的信息
        this.then = (callback) => {
            // 状态是成功的时候, 执行回调
            if (this.status === 'fulfilled') {
                callback(this.sucessInfo);
            }
            return this;
        };
        // 处理失败的信息
        this.catch = (callback) => {
            if (this.status === 'rejected') {
                callback(this.errorInfo);
            }
            return this;
        };

        // 将成功信息保存到sucessInfo
        this.resolve = (data) => {
            // 改变状态为成功
            this.status = 'fulfilled';
            this.sucessInfo = data;
        };
        // 将失败信息保存到errorInfo
        this.reject = (data) => {
            // 改变状态为成功
            this.status = 'rejected';
            this.errorInfo = data;
        };

        callback(this.resolve, this.reject);
    }

    const promise = new MyPromise((resolve, reject) => {
        let flag = true;
        if (flag) {
            resolve('我成功了, 哈哈哈哈哈哈');
        } else {
            reject('失败了, 继续努力');
        }
    });


    promise.then((res) => {
        console.log(res);
    }).catch((err) => {
        console.log(err);
    })
</script>
```

### (3) 手写promise终版

#### 01 class

```javascript
class Cat{
  constructor(name,age) {
    this.name = name;
    this.age = age;
    this.hobby = '抓老鼠';
  }

  say() {
    console.log(this.name)
  }
}

var cat = new Cat('小花',2);
console.log(cat)
```

#### 02 bind

```javascript
<script>
    class Cat {
        constructor(name, age) {
            this.name = name;
            this.age = age;
            this.hobby = '抓老鼠';
            // 让say方法里的this永远指向创建出来的实例
            this.say = this.say.bind(this);
        }

        say() {
            console.log(this);
        }
    }

    var cat = new Cat('小花', 2);
    var say = cat.say;
    say(); 
</script>
```

#### 03 发布订阅模式(略)

#### 04 实现promise

```javascript
<script>
    class MyPromise {
        constructor(callback) {
            // 定义状态
            this.status = "pending";
            // 收集订阅者的信息
            this.info = [];
            // 存放then方法回调函数的返回值
            this.successInfo = null;

            // 让this永远指向实例
            this.resolve = this.resolve.bind(this);
            this.reject = this.reject.bind(this);

            callback(this.resolve, this.reject);
        }

        // 订阅成功信息
        then(callback) {
            // 把订阅信息存入this.info
            this.info.push({
                status: "fulfilled",
                callback
            });
            return this;
        }

        // 订阅失败信息
        catch (callback) {
            this.info.push({
                status: "rejected",
                callback
            });
        }

        // 发布成功的信息
        resolve(data) {
            this.status = "fulfilled";
            // 把this.info里订阅了"成功"的订阅者取出来
            let successList = this.info.filter((item) => item.status === "fulfilled");
            // 执行回调, 通知订阅者
            successList.forEach((item) => {
                if (!this.successInfo) {
                    // 第一次调用
                    this.successInfo = item.callback(data);
                } else {
                    this.successInfo = item.callback(this.successInfo);
                }
            });
        }

        // 发布失败的信息
        reject(data) {
            this.status = "fulfilled";
            // 把this.info里订阅了"成功"的订阅者取出来
            let errorList = this.info.filter((item) => item.status === "rejected");
            // 执行回调, 通知订阅者
            errorList.forEach((item) => {
                item.callback(data);
            });
        }
    }

    const promise = new MyPromise((resolve, reject) => {
        let flag = true;
        setTimeout(() => {
            if (flag) {
                resolve("我成功了");
            } else {
                reject("失败了, 继续努力");
            }
        }, 1000);
    });

    // then和catch是订阅, then订阅是成功的信息, catch订阅的是失败的信息
    promise.then((res) => {
        console.log(res);
        return res + 1;
    }).then(res => {
        console.log(res);
        return res + 1;
    }).then(res => {
    	console.log(res);
    }).catch((err) => {
        console.log(err);
    });
</script>  
```

原文地址:  [https://segmentfault.com/a/1190000023180502](https://segmentfault.com/a/1190000023180502)

### (4) promise.all作用   


1. Promise.all(iterable) 方法返回一个 Promise 实例，此实例在 iterable 参数内所有的 promise 都“完成（resolved）”或
   参数中不包含 promise 时回调完成（resolve）；如果参数中  promise 有一个失败（rejected），此实例回调失败（reject），
   失败的原因是第一个失败 promise 的结果。
1. 它通常在启动多个异步任务并发运行并为其结果创建承诺之后使用，以便人们可以等待所有任务完成。
1. 通俗的说法：解决promise的同步调用问题，可以将多个异步操作转为同步操作，缺点：只要有一个promise返回错误，那么后面的promise都不会再执行



```javascript
let p1 = new Promise((resolve,reject) => {
  setTimeout(()=> {
    resolve({name:'张三'});
  },1000)
})
let p2 = new Promise((resolve,reject) => {
  setTimeout(()=> {
    reject({name:'李四'});
  },2000)
})
let p3 = new Promise((resolve,reject) => {
  setTimeout(()=> {
    resolve({name:'王五'});
  },3000)
})

Promise.all([p1,p2,p3]).then(res=> {
  console.log('成功信息',res)
}).catch(err=> {
  console.log('失败信息',err);
});
```





## 作业: 

1. 完成每日生鲜搜索功能

2. 说出以下输出结果, 并解释为什么

   ```JavaScript
   function fun() {
       var r = new Array();
       for(var i=0;i<3;i++) {
           r[i] = function() {
               return i;
           }
       }
       return r;
   }
   var fun2 = fun();
   for(var i=0;i<fun2.length;i++) {
       console.log(fun2[i]()); 
   }
   ```

3. 说出以下代码运行结果, 并解释为什么

   ```JavaScript
   var name = 'xiao ming';
   var obj = {
     name: 'xiao hong',
     getName: function() {
       return function() {
         return this.name;
       }
     }
   } 
   console.log(obj.getName()());
   ```

4. 说出以下代码运行结果, 并解释为什么

   ```JavaScript
   var param = 1;
   function main() { 
     console.log('1',param);
     var param = 2;
     console.log('2',this.param);
     this.param = 3;
   }
   
   main(); 
   var m = new main(); 
   ```

   
